/*
 *  -- MAGMA (version 1.2.0) --
 *     Univ. of Tennessee, Knoxville
 *     Univ. of California, Berkeley
 *     Univ. of Colorado, Denver
 *     @date September 2014
 *
 * @generated from testing_ztrmv.cpp normal z -> d, Wed Sep 17 15:37:46 2014
 * @author Chongxiao Cao
 *
 **/
// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "magma_types.h"
#include "testings.h"


/* ////////////////////////////////////////////////////////////////////////////
   -- Testing dtrmv
*/
int main( int argc, char** argv)
{
    /* Initialize */
    magma_queue_t  queue;
    magma_device_t device;
    magma_int_t num = 0;
    magma_int_t err;
    magma_init();
    err = magma_get_devices( &device, 1, &num );
    if ( err != MAGMA_SUCCESS or num < 1 ) {
      fprintf( stderr, "magma_get_devices failed: %d\n", (int) err );
      exit(-1);
    }
    err = magma_queue_create( device, &queue );
    if ( err != MAGMA_SUCCESS ) {
      fprintf( stderr, "magma_queue_create failed: %d\n", (int) err );
      exit(-1);
    }

    real_Double_t   gflops, mic_perf, mic_time, cpu_perf, cpu_time;
    double          mic_error, Cnorm, work[1];
    magma_int_t N;
    magma_int_t Ak;
    magma_int_t sizeA;
    magma_int_t lda, ldda;
    magma_int_t ione     = 1;
    magma_int_t ISEED[4] = {0,0,0,1};
    
    double *h_A, *h_x, *h_xmic;
    double *d_A, *d_x;
    double c_neg_one = MAGMA_D_NEG_ONE;
    
    magma_opts opts;
    parse_opts( argc, argv, &opts );
    
    printf("    N   MIC Gflop/s (ms)   CPU Gflop/s (ms)  MIC error\n");
    printf("==================================================================\n");
    for( int i = 0; i < opts.ntest; ++i ) {
        for( int iter = 0; iter < opts.niter; ++iter ) {
            N = opts.nsize[i];
            gflops = FLOPS_DTRMM(opts.side, N, 1) / 1e9;

            lda = N;
            Ak = N;
            
            ldda = ((lda+31)/32)*32;
            
            sizeA = lda*Ak;
            
            TESTING_MALLOC_HOST( h_A,  double, lda*Ak );
            TESTING_MALLOC_HOST( h_x,  double, N  );
            TESTING_MALLOC_HOST( h_xmic, double, N  );
            
            TESTING_MALLOC_DEV( d_A, double, ldda*Ak );
            TESTING_MALLOC_DEV( d_x, double, N  );
            
            /* Initialize the matrices */
            lapackf77_dlarnv( &ione, ISEED, &sizeA, h_A );
            lapackf77_dlarnv( &ione, ISEED, &N, h_x );
            
            /* =====================================================================
               Performs operation using CUDA-BLAS
               =================================================================== */
            magma_dsetmatrix( Ak, Ak, h_A, 0, lda, d_A, 0, ldda, queue );
            //magma_dsetvector( N, h_x, 1, d_x, 1 );
            magma_dsetmatrix( N, 1, h_x, 0, N, d_x, 0, N, queue );
            
            mic_time = magma_sync_wtime( NULL );
            magma_dtrmv( opts.uplo, opts.transA, opts.diag,
                         N,
                         d_A, 0, ldda,
                         d_x, 0, 1, queue );
            mic_time = magma_sync_wtime( NULL ) - mic_time;
            mic_perf = gflops / mic_time;
            
            
            /* =====================================================================
               Performs operation using CPU BLAS
               =================================================================== */
            if ( opts.check ) {
                //magma_dgetvector( N, d_x, 1, h_xmic, 1 );
                magma_dgetmatrix( N, 1, d_x, 0, N, h_xmic, 0, N, queue );
                cpu_time = magma_wtime();
                blasf77_dtrmv( lapack_const(opts.uplo), lapack_const(opts.transA), lapack_const(opts.diag),
                                &N,
                                h_A, &lda,
                                h_x, &ione );
                cpu_time = magma_wtime() - cpu_time;
                cpu_perf = gflops / cpu_time;

                /* =====================================================================
                   Check the result
                   =================================================================== */
                // compute relative error for both magma & mic, relative to lapack,
                // |C_magma - C_lapack| / |C_lapack|
                Cnorm = lapackf77_dlange( "M", &N, &ione, h_x, &N, work );

                blasf77_daxpy( &N, &c_neg_one, h_x, &ione, h_xmic, &ione );
                mic_error = lapackf77_dlange( "M", &N, &ione, h_xmic, &N, work ) / Cnorm;

                printf("%5d   %7.2f (%7.2f)   %7.2f (%7.2f)    %8.2e\n",
                        (int) N,
                        mic_perf, 1000.*mic_time,
                        cpu_perf,    1000.*cpu_time,
                        mic_error );
            }
            else {
                printf("%5d   %7.2f (%7.2f)    ---   (  ---  )    ---\n",
                        (int) N,
                        mic_perf, 1000.*mic_time);
            }

            TESTING_FREE_HOST( h_A  );
            TESTING_FREE_HOST( h_x  );
            TESTING_FREE_HOST( h_xmic );

            TESTING_FREE_DEV( d_A );
            TESTING_FREE_DEV( d_x );
        }
        if ( opts.niter > 1 ) {
            printf( "\n" );
        }
    }

    magma_queue_destroy( queue );
    magma_finalize();

    return 0;
}
