/*
    -- MAGMA (version 1.2.0) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date September 2014

       @generated from testing_ztrsv.cpp normal z -> d, Wed Sep 17 15:37:46 2014
*/

// make sure that asserts are enabled
#ifdef NDEBUG
#undef NDEBUG
#endif
#include <assert.h>

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, project
#include "flops.h"
#include "magma.h"
#include "magma_lapack.h"
#include "magma_types.h"
#include "testings.h"

#define h_A(i,j) (h_A + (i) + (j)*lda)

/* ////////////////////////////////////////////////////////////////////////////
   -- Testing dtrsm
*/
int main( int argc, char** argv)
{
    /* Initialize */
    magma_queue_t  queue;
    magma_device_t device;
    magma_int_t num = 0;
    magma_int_t err;
    magma_init();
    err = magma_get_devices( &device, 1, &num );
    if ( err != MAGMA_SUCCESS or num < 1 ) {
      fprintf( stderr, "magma_get_devices failed: %d\n", (int) err );
      exit(-1);
    }
    err = magma_queue_create( device, &queue );
    if ( err != MAGMA_SUCCESS ) {
      fprintf( stderr, "magma_queue_create failed: %d\n", (int) err );
      exit(-1);
    }

    real_Double_t   gflops, mic_perf, mic_time, cpu_perf=0, cpu_time=0;
    double          mic_error, normA, normx, normr, work[1];
    magma_int_t N, info;
    magma_int_t sizeA;
    magma_int_t lda, ldda;
    magma_int_t ione     = 1;
    magma_int_t ISEED[4] = {0,0,0,1};
    magma_int_t *ipiv;

    double *h_A, *h_b, *h_x, *h_xmic;
    double *d_A, *d_x;
    double c_neg_one = MAGMA_D_NEG_ONE;
    
    magma_opts opts;
    parse_opts( argc, argv, &opts );
    
    printf("uplo = %s, transA = %s, diag = %s\n",
           lapack_const(opts.uplo), lapack_const(opts.transA), lapack_const(opts.diag) );
    printf("    N  MIC Gflop/s (ms)   CPU Gflop/s (ms)   MIC error\n");
    printf("============================================================\n");
    for( int itest = 0; itest < opts.ntest; ++itest ) {
        for( int iter = 0; iter < opts.niter; ++iter ) {
            N = opts.nsize[itest];
            gflops = FLOPS_DTRSM(opts.side, N, 1) / 1e9;
            lda    = N;
            ldda   = ((lda+31)/32)*32;
            sizeA  = lda*N;
            
            TESTING_MALLOC_HOST( ipiv,      magma_int_t,        N     );
            TESTING_MALLOC_HOST( h_A,       double, lda*N );
            TESTING_MALLOC_HOST( h_b,       double, N     );
            TESTING_MALLOC_HOST( h_x,       double, N     );
            TESTING_MALLOC_HOST( h_xmic, double, N     );
            
            TESTING_MALLOC_DEV( d_A, double, ldda*N );
            TESTING_MALLOC_DEV( d_x, double, N      );
            
            /* Initialize the matrices */
            /* Factor A into LU to get well-conditioned triangular matrix.
             * Copy L to U, since L seems okay when used with non-unit diagonal
             * (i.e., from U), while U fails when used with unit diagonal. */
            lapackf77_dlarnv( &ione, ISEED, &sizeA, h_A );
            lapackf77_dgetrf( &N, &N, h_A, &lda, ipiv, &info );
            for( int j = 0; j < N; ++j ) {
                for( int i = 0; i < j; ++i ) {
                    *h_A(i,j) = *h_A(j,i);
                }
            }
            
            lapackf77_dlarnv( &ione, ISEED, &N, h_b );
            blasf77_dcopy( &N, h_b, &ione, h_x, &ione );
            
            /* =====================================================================
               Performs operation using MKL
               =================================================================== */
            magma_dsetmatrix( N, N, h_A, 0, lda, d_A, 0, ldda, queue );
            //magma_dsetvector( N, h_x, 1, d_x, 1 );
            magma_dsetmatrix( N, 1, h_x, 0, 1, d_x, 0, 1, queue );
            
            mic_time = magma_sync_wtime( NULL );
            magma_dtrsv( opts.uplo,
                         opts.transA, opts.diag,
                         N,
                         d_A, 0, ldda,
                         d_x, 0, 1, queue );
            mic_time = magma_sync_wtime( NULL ) - mic_time;
            mic_perf = gflops / mic_time;
            
            //magma_dgetvector( N, d_x, 1, h_xmic, 1 );
            magma_dgetmatrix( N, 1, d_x, 0, 1, h_xmic, 0, 1, queue );
            
            /* =====================================================================
               Performs operation using CPU BLAS
               =================================================================== */
            if ( opts.lapack ) {
                cpu_time = magma_wtime();
                blasf77_dtrsv( lapack_const(opts.uplo), lapack_const(opts.transA), lapack_const(opts.diag),
                               &N,
                               h_A, &lda,
                               h_x, &ione );
                cpu_time = magma_wtime() - cpu_time;
                cpu_perf = gflops / cpu_time;
            }
            
            /* =====================================================================
               Check the result
               =================================================================== */
            // ||b - Ax|| / (||A||*||x||)
            // error for MKL
            normA = lapackf77_dlange( "F", &N, &N, h_A, &lda, work );
            
            normx = lapackf77_dlange( "F", &N, &ione, h_xmic, &ione, work );
            blasf77_dtrmv( lapack_const(opts.uplo), lapack_const(opts.transA), lapack_const(opts.diag),
                           &N,
                           h_A, &lda,
                           h_xmic, &ione );
            blasf77_daxpy( &N, &c_neg_one, h_b, &ione, h_xmic, &ione );
            normr = lapackf77_dlange( "F", &N, &ione, h_xmic, &N, work );
            mic_error = normr / (normA*normx);

            if ( opts.lapack ) {
                printf("%5d   %7.2f (%7.2f)   %7.2f (%7.2f)   %8.2e\n",
                        (int) N,
                        mic_perf, 1000.*mic_time,
                        cpu_perf,    1000.*cpu_time,
                        mic_error );
            }
            else {
                printf("%5d   %7.2f (%7.2f)     ---  (  ---  )   %8.2e\n",
                        (int) N,
                        mic_perf, 1000.*mic_time,
                        mic_error );
            }
            
            TESTING_FREE_HOST( ipiv );
            TESTING_FREE_HOST( h_A  );
            TESTING_FREE_HOST( h_b  );
            TESTING_FREE_HOST( h_x  );
            TESTING_FREE_HOST( h_xmic );
            
            TESTING_FREE_DEV( d_A );
            TESTING_FREE_DEV( d_x );
        }
        if ( opts.niter > 1 ) {
            printf( "\n" );
        }
    }

    magma_queue_destroy( queue );
    magma_finalize();

    return 0;
}
